//********************************************************************************
//* File       : WaylandCB.cpp                                                   *
//* Author     : Mahlon R. Smith                                                 *
//*              Copyright (c) 2019-2023 Mahlon R. Smith, The Software Samurai   *
//*                  GNU GPL copyright notice located below                      *
//* Date       : 25-Apr-2020                                                     *
//* Version    : (see WaylandCB_Version below)                                   *
//*                                                                              *
//* Description:                                                                 *
//* The WaylandCB class provides a simple interface between console              *
//* applications and the GNOME/Wayland compositor's two clipboards:              *
//* "Primary" and "Regular".                                                     *
//*                                                                              *
//* There are two modes of access: "indirect" and "direct".                      *
//* [At this time, WaylandCB does not implement the "direct" method.]            *
//* -- The "indirect" method relies upon the wl-clipboard utilities              *
//*    "wl-copy" and "wl-paste".                                                 *
//* -- The "direct" method uses the same method for accessing the clipboards     *
//*    that the wl-copy/wl-paste utilities use, but bypasses the utilities       *
//*    and interfaces directly with GNOME/Wayland (presumably through the        *
//*    GTK+ toolkit).                                                            *
//*                                                                              *
//* Only plain text data (UTF-8 or UTF-32) may be placed on the clipboard or     *
//* retrieved from the clipboard. Null-termination of the source data for the    *
//* copy operation is HIGHLY RECOMMENDED, but not enforced.                      *
//* Copy-and-paste of binary data is not supported.                              *
//*                                                                              *
//*                                                                              *
//* Development Tools:                                                           *
//*   Fedora Linux 30, GNOME terminal 3.32                                       *
//*   Wintel (Lenovo) hardware.                                                  *
//*   GNU G++ (Gcc v:9.2.1)                                                      *
//********************************************************************************
//* Copyright Notice:                                                            *
//* This program is free software: you can redistribute it and/or modify it      *
//* under the terms of the GNU General Public License as published by the Free   *
//* Software Foundation, either version 3 of the License, or (at your option)    *
//* any later version.                                                           *
//*                                                                              *
//* This program is distributed in the hope that it will be useful, but          *
//* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY   *
//* or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License     *
//* for more details.                                                            *
//*                                                                              *
//* You should have received a copy of the GNU General Public License along      *
//* with this program.  If not, see <http://www.gnu.org/licenses/>.              *
//*                                                                              *
//*         Full text of the GPL License may be found in the Texinfo             *
//*         documentation for this program under 'Copyright Notice'.             *
//********************************************************************************
//* Version History (most recent first):                                         *
//*                                                                              *
//* v: 0.0.03 18-Apr-2020                                                        *
//*   -- Redefine the return value for the wcbTest() method.                     *
//*      Formerly, it returned 'true' indicating a successful connection,        *
//*      or 'false' if unable to communicate with the system clipboard.          *
//*      The return value is now a member of enum wcbStatus which provides a     *
//*      more detailed status.                                                   *
//*   -- Redefine the return value for the wcbClear() method.                    *
//*      Formerly, it returned (0) on success or (-1) if error.                  *
//*      The return value is now 'true' for success or 'false' if error.         *
//*   -- Cosmetic changes: Update some source code comments.                     *
//*   -- Update documentation.                                                   *
//*                                                                              *
//* v: 0.0.02 12-Jan-2020                                                        *
//*   -- All functionality verified.                                             *
//*   -- Addition of the delay before every access to the external utilities     *
//*      has significantly reduced the communications failure rate.              *
//*   -- Integrated into the NcDialogAPI test application "Dialogx".             *
//*   -- Demonstration program, "Wayclip" is now a pure console application.     *
//*   -- Both "Dialogx" and "Wayclip" exercise the full functionality of         *
//*      the WaylandCB class.                                                    *
//*   -- Update documentation.                                                   *
//*                                                                              *
//* v: 0.0.01 29-Dec-2019                                                        *
//*   -- First effort: Built as part of the experimental NcDialog utility        *
//*      "Wayclip" which is modelled on the "Dialogx" utility previously         *
//*      designed as a way to access the X-clipboard.                            *
//*      When this experimental application is stable, it will either be         *
//*      integrated into OR replace the "Dialogx" application.                   *
//*   -- Test command-line access to the Wayland clipboard using the             *
//*      wl-clipboard pair of utilities: "wl-copy" and "wl-paste".               *
//*      These are simple (possibly too simple) utilities that communicate       *
//*      between the terminal and the Wayland clipboards.                        *
//*      Note that these utilities may not be installed by default, even         *
//*      on a GNOME/Wayland desktop. If invoked on a system where they are       *
//*      not installed, yum/dnf offers to install them. (/usr/bin)               *
//*   -- Temp-file manipulation is taken from our FileMangler application and    *
//*      simplified for this application.                                        *
//*   -- Escaping of "special" characters in clipboard data is borrowed from     *
//*      the FileMangler algorithm used to escape characters in filespecs.       *
//*      This escaping is necessary for passing commands to the shell program.   *
//*      (A much-simplified version of escaping is currently used.)              *
//*   -- "UTF8_STRING" is an outgrowth of X11 support for UTF-8 encoding an is   *
//*      an "official" X11 "atom" (token). It seems that the Wayland             *
//*      developers have embraced it as a proper representation for UTF-8        *
//*      encoding. The wl-clipboard sees subtle differences between              *
//*      "text/plain;charset=UTF-8" and its shortcut "UTF8_STRING".              *
//*      For our purposes, "text/plain;charset=UTF-8" is preferred and either    *
//*      it or its shortcut "UTF8_STRING" are used exclusively within this       *
//*      class when specifying the MIME type for an operation.                   *
//*                                                                              *
//*                                                                              *
//*                                                                              *
//********************************************************************************
//* NOTES:                                                                       *
//* ------                                                                       *
//* The wl-clipboard documentation is simplistic at best; in fact it is a        *
//* "man page", not even a Texinfo document. These utilities do not even         *
//* report a '--version' or '--help' option.                                     *
//* Here it is in its entirety (v:2018-10-03):                                   *
//*                                                                              *
//* NAME                                                                         *
//*    wl-clipboard - Wayland copy and paste command line utiltites              *
//*                                                                              *
//* SYNOPSIS                                                                     *
//*    wl-copy [--primary] [--paste-once] [--foreground] [--clear]               *
//*            [--type mime/type] [text...]                                      *
//*                                                                              *
//*    wl-paste [--primary] [--no-newline] [--list-types] [--type mime/type]     *
//*                                                                              *
//* DESCRIPTION                                                                  *
//*    wl-copy copies the given text to the Wayland clipboard.                   *
//*    If no text is given, wl-copy copies data from its standard input.         *
//*                                                                              *
//*    wl-paste pastes data from the Wayland clipboard to its standard output.   *
//*                                                                              *
//*    Although  wl-copy  and  wl-paste are particularly optimized for plain     *
//*    text and other textual content formats, they fully support content of     *
//*    arbitrary MIME types. wl-copy automatically infers the type of the        *
//*    copied content by running xdg-mime(1) on it. wl-paste tries its best to   *
//*    pick a type to paste based on the list of offered MIME types and the      *
//*    extension of the file it's pasting into. If you're not satisfied with     *
//*    the type they pick or don't want to rely on this implicit type            *
//*    inference, you can explicitly specify the type to use with the            *
//*    --type option.                                                            *
//*                                                                              *
//* OPTIONS                                                                      *
//* -p, --primary                                                                *
//*     Use the "primary" clipboard instead of the regular clipboard.            *
//*     This uses the private GTK+ primary selection protocol.                   *
//*     See the GNOME  Wiki page on primary selection under Wayland              *
//*     ⟨https://wiki.gnome.org/Initiatives/Wayland/PrimarySelection⟩ for more   *
//*     details.                                                                 *
//*                                                                              *
//* -o, --paste-once                                                             *
//*     Only serve one paste request and then exit. Unless a clipboard manager   *
//*     specifically designed to prevent this is in use, this has the effect     *
//*     of clearing the clipboard after the first paste, which is useful for     *
//*     copying sensitive data such as passwords. Note that this may break       *
//*     pasting into some clients, in particular pasting into XWayland windows   *
//*     is known to break when this option is used.                              *
//*                                                                              *
//* -f, --foreground                                                             *
//*     By  default, wl-copy forks and serves data requests in the background;   *
//*     this option overrides that behavior, causing wl-copy to run in the       *
//*     foreground.                                                              *
//*                                                                              *
//* -c, --clear                                                                  *
//*     Instead of copying anything, clear the clipboard so that nothing is      *
//*     copied.                                                                  *
//*                                                                              *
//* -n, --no-newline                                                             *
//*     Do not append a newline character after the pasted clipboard content.    *
//*     This option is automatically enabled for non-text content types.         *
//*                                                                              *
//* -t mime/type, --type mime/type                                               *
//*     Override the automatically selected MIME type. For wl-copy this option   *
//*     controls which type wl-copy will offer the content as. For wl-paste      *
//*     it  controls  which  of  the offered types wl-paste will request the     *
//*     content in. In addition to specific MIME types such as image/png,        *
//*     wl-paste also accepts generic type names such as text and image which    *
//*     make it automatically pick some offered MIME type that matches the       *
//*     given generic name.                                                      *
//*                                                                              *
//* -l, --list-types                                                             *
//*     Instead of pasting the selection, output the list of MIME types it       *
//*     is offered in.                                                           *
//*                                                                              *
//* ENVIRONMENT                                                                  *
//*   WAYLAND_DISPLAY                                                            *
//*   Specifies what Wayland server wl-copy and wl-paste should connect to.      *
//*   This is the same environment variable that you pass to other Wayland       *
//*   clients, such as graphical aplications, that connect to this Wayland       *
//*   server. It is normally set up automatically by the graphical  session      *
//*   and the Wayland compositor. See wl_display_connect(3) for more details.    *
//*   WAYLAND_DEBUG                                                              *
//*   When  set  to  1,  causes  the wayland-client(7) library to log every      *
//*   interaction wl-copy and wl-paste make with the Wayland compositor to       *
//*    stderr.                                                                   *
//*                                                                              *
//* EXAMPLES                                                                     *
//*   $ wl-copy Hello world!                                                     *
//*   $ ls ~/Downloads | wl-copy                                                 *
//*   $ wl-copy < ~/Pictures/photo.png                                           *
//*   $ wl-copy "!!"                                                             *
//*   $ wl-paste -n > clipboard.txt                                              *
//*   $ wl-paste --list-types | wl-copy                                          *
//* SEE ALSO: xclip(1)                                                           *
//*                                                                              *
//*                                                                              *
//* ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  - *
//* Analysis:                                                                    *
//* ---------                                                                    *
//* wl-paste: unrecognized (invalid) options:                                    *
//*           --clear('-c')                                                      *
//*           --foreground('-f')                                                 *
//*           --paste-once('-o')                                                 *
//* wl-paste: --primary('-p')  is functional:                                    *
//*           With it the "primary" clipboard contents are returned.             *
//*           Without it the "regular" clipboard contents are returned.          *
//* wl-paste: --no-newline('-n') is functional                                   *
//* wl-paste: --list-types('-l') is functional                                   *
//*           For our purposes, "text/plain;charset=UTF-8" is preferred and      *
//*           either it or its shortcut "UTF8_STRING" are used exclusively       *
//*           within this class.                                                 *
//* wl-paste: --type('-t') is recognized and it tentatively presumed to be       *
//*           functional.                                                        *
//*           ---  ---  ---  ---  ---  ---  ---  ---  ---                        *
//* wl-copy:  unrecognized (invalid) options:                                    *
//*           --no-newline('-n')                                                 *
//*           --list-types('-l')                                                 *
//* wl-copy:  --primary('-p') is functional                                      *
//*           With it the specified data are copied to the "primary" clipboard.  *
//*           Without it, the specified data are copied to the "regular"         *
//*           clipboard.                                                         *
//* wl-copy:  --paste-once('-o') is functional                                   *
//*           When 'wl-paste' is called the first time, the specified data are   *
//*           returned                                                           *
//*           When 'wl-paste' is called a second time (or multiple additional    *
//*           times), the message: "No selection" is returned.                   *
//*           NOTE: This represents a _logical_ error because it _appears_       *
//*           to place the literal string "No selection" on the clipboard        *
//*           although it _actually_ does not (clipboard is empty).              *
//* wl-copy:  --foreground('-f') is inferred to be functional because it does    *
//*           not release the command-line process until CTRL+C.                 *
//* wl-copy:  --clear('-c') is recognized, but apparently does nothing.          *
//*           Instead, send an empty string to clear the clipboard.              *
//* wl-copy:  --type('-t') appears to be functional.                             *
//*           If UTF8_STRING is specified along with the source text, then the   *
//*           range of types returned is restricted to those which overlap       *
//*           (e.g. UTF-16 is excluded).                                         *
//*           NOTE: This option has some logical problems:                       *
//*           1) Subsequent wl-copy commands will retain the previous value.     *
//*              An external copy-to-clipboard will reset to the full list.      *
//*           2) -t=XXX is not parsed properly and includes the '=' as part      *
//*              of the type specification.                                      *
//*           3) If a known type is specified, then when the available types     *
//*              are requested then the specified type will be reported twice.   *
//*                                                                              *
//* The utilities are fragile. If there is a program error, (and there were      *
//* many during development), then we had to go to the command line and          *
//* call the utilities to directly copy/paste something before it would          *
//* respond again within the application. Presumably this is because the         *
//* process that controls them was left in an indeterminate state by our         *
//* error in setting the clipboard contents. Note however, that wl-clipboard     *
//* does generate an "Application Crash" message for the system:                 *
//*       wl-clipboard quit unexpectedly                                         *
//*       The application encountered a problem and could not continue.          *
//*       Version: 1.0.0-0.2.20181215git7e5103c.fc30.x86_64                      *
//*       /var/spool/abrt/ccpp-2020-01-05-12:11:59.873358-11380                  *
//* This should show up in the desktop's crash reports.                          *
//*                                                                              *
//* Note also that these utilities are unbearably slow. If commands are sent     *
//* in quick succession, it can easily generate a system exception, either       *
//* by the utilities OR by the shell program.                                    *
//*          Example:  free(): double free detected in tcache 2                  *
//*                                                                              *
//* For these reasons, we have implemented a small pause before executing        *
//* a call to either 'wl-copy' or 'wl-paste' (see Systemcall() method).          *
//* This may not fully compensate for the weaknesses in these utilities, but     *
//* it does seem to reduce the number of errors.                                 *
//*                                                                              *
//********************************************************************************

//************
//* Includes *
//************
#include "WaylandCB.hpp"

//*********************
//* Local Definitions *
//*********************
#define ZERO (0)
#define NULLCHAR ('\0')       // NULL terminator character
#define TILDE ('~')           // Highest ASCII character code
#define SPACE (' ')           // ASCII space character
#define DSPACE (0x3000)       // CJK 2-column space character
#define TAB ('\t')            // Horizontal TAB character
#define VTAB ('\v')           // Vertical TAB character
#define NEWLINE ('\n')        // Newline character
#define CR ('\r')             // Carriage-return character
#define FF ('\f')             // Form-feed character
typedef struct stat64 FileStats ;

//*****************
//* Constant Data *
//*****************
static const char* const WaylandCB_Version = "0.0.03" ;   // class version

//* Except for development/debugging, this is the only      *
//* MIME-type/encoding used within this class, and except   *
//* for legacy data (and Windoze) this is the only          *
//* acceptable format for plain-text data.                  *
static const char* const dfltMIME = "text/plain;charset=UTF-8" ;
static const short dfltMIME_LEN = 24 ;
//* This is an X11 shortcut for "text/plain;charset=UTF-8". *
static const char* const dfltX11 = "UTF8_STRING" ;
static const short dfltX11_LEN = 11 ;
static const short PADCNT = 32 ;    // padding added to conversion buffer for escaped chars

//*************************
//* Non-member Prototypes *
//*************************


//*************************
//*      ~WaylandCB       *
//*************************
//******************************************************************************
//* Destructor:                                                                *
//*  1) Delete all temp files.                                                 *
//*  2) Return all resources to the system                                     *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

WaylandCB::~WaylandCB ( void )
{
   this->DeleteTempname ( this->tfspec ) ;

}  //* End ~WaylandCB() *

//*************************
//*       WaylandCB       *
//*************************
//******************************************************************************
//* Default Constructor:                                                       *
//*  1) Initialize all class data members.                                     *
//*  2) Locate the system's temp-file directory.                               *
//*  3) Create a temporary file for communicating with the clipboard interface.*
//*  4) Determine whether "wl-copy" and "wl-paste" utilities are installed.    *
//*  5) Test access to the Wayland clipboard.                                  *
//*                                                                            *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: constructors implicitly return a pointer to the object            *
//******************************************************************************

WaylandCB::WaylandCB ( void )
{

   this->reset () ;                 // initialize all data members

   #if WCB_DEBUG != 0
   //* Create a debugging log.*
   this->dbofs.open( "./wcblog.txt", ofstream::out | ofstream::trunc ) ;
   if ( this->dbofs.is_open() )
      this->dbofs << "WaylandCB Debug Log\n===================" << endl ;
   #endif   // WCB_DEBUG

   //* Execute the startup sequence.*
   this->StartupSequence () ;

}  //* End WaylandCB() *

//*************************
//*        wcbGet         *
//*************************
//******************************************************************************
//* Public Method                                                              *
//* -------------                                                              *
//* Get a copy of the Wayland clipboard data.                                  *
//* Data returned will be NULL terminated, no trailing newline.                *
//*                                                                            *
//* Input  : trg     : pointer to buffer to receive data                       *
//*                    wchar_t*  or  char*  or  gString&                       *
//*          primary : (optional, 'false' by default)                          *
//*                    'false', read from "regular" (ordinary, main) clipboard *
//*                    'true', read from "primary" (highlighted-data) clipboard*
//*          len     : (optional) specify length of target buffer              *
//*                      for wchar_t* targets, default == gsMAXCHARS           *
//*                      for char* targets, default == gsMAXBYTES              *
//*                      for gString targets, default == gsMAXBYTES            *
//*                    Important Note: Longer buffers may be specified for     *
//*                      char* targets; but for wchar_t* and gString objects,  *
//*                      this is the max length because it is the size of the  *
//*                      UTF-8 to UTF-32 conversion buffer.                    *
//*                                                                            *
//* Returns: for wchar_t target, number of characters read (incl.NULLCHAR)     *
//*          for char target or gString target, number of bytes read           *
//*           (incl. NULLCHAR)                                                 *
//*          returns wcbsNOCONNECT if error i.e. not connected to clipboard    *
//*          returns wcbsNOINSTALL if 'wl-copy' utility not installed          *
//******************************************************************************
//* Notes:                                                                     *
//* 1) If 'wl-paste' is not installed, there will be a message simiar to:      *
//*          "sh: wl-paste: command not found"                                 *
//*                                                                            *
//*                                                                            *
//******************************************************************************

short WaylandCB::wcbGet ( char* trg, bool primary, short len )
{
   const char* const getTemplate = "wl-paste %s -nt \"%s\" 1>>\"%s\" 2>>\"%s\"" ;
   short byteCount = ZERO ;   // return value

   *trg = NULLCHAR ;          // initialize caller's buffer

   //* If not connected to the Wayland clipboard *
   //* interface, return an error condition.     *
   if ( ! this->connected )
      return ( wcbsNOCONNECT ) ;

   //* Discard contents (if any) of the target temp file.*
   ofstream ofs( this->tfspec, ofstream::out | ofstream::trunc ) ;
   if ( ofs.is_open() )
      ofs.close() ;

   //* Determine the size of the command buffer *
   gString gs( getTemplate ) ;
   short buffLen = gs.utfbytes() + dfltX11_LEN + (this->tfspecLen * 2) + PADCNT ;
   char cmdBuff[buffLen] ;

   //* Construct the command *
   snprintf ( cmdBuff, buffLen, getTemplate, 
              (char*)(primary ? "-p" : " "), dfltX11, this->tfspec, this->tfspec ) ;

   #if WCB_DEBUG != 0
   if ( this->dbofs.is_open() )
   {
      this->gsdb.compose( "-- wcbGet( buffLen:%hd )\n   %s", &buffLen, cmdBuff ) ;
      this->dbofs << this->gsdb.ustr() << endl ;
   }
   #endif   // WCB_DEBUG

   //* Capture the clipboard *
   this->Systemcall ( cmdBuff ) ;

   //* Retrieve the results from the temp file.*
   ifstream ifs( this->tfspec, ifstream::in ) ; // open the file
   if ( ifs.is_open() )                         // if file opened
   {
      char inByte ;                             // input buffer
      while ( byteCount < (len - 1) )           // loop
      {
         ifs.read( &inByte, 1 ) ;               // read a byte
         if ( ifs.good() || (ifs.gcount() == 1) ) // if read successful
            trg[byteCount++] = inByte ;         // save the byte
         else                                   // end-of-file
            break ;
      }
      //* Discard any trailing newline and terminate the string.*
      if ( (byteCount > ZERO) && (trg[byteCount - 1] == NEWLINE) )
         --byteCount ;
      if ( (byteCount == ZERO) || (trg[byteCount - 1] != NULLCHAR) )
         trg[byteCount++] = NULLCHAR ;
      ifs.close() ;                             // close the file

      //* Test for a "wl-paste not found" message.*
      gs = trg ;
      if ( ((gs.find( L"wl-paste" )) >= ZERO ) &&
           ((gs.find( L"command not found" )) >= ZERO) )
      {
         // Programmer's Note: This should never happen because on startup,
         // 'wl-copy' is called first, and if it fails, this method should 
         // never be called. This assumes that if 'wl-copy' is installed 
         // then 'wl-paste' will also be installed, but note that our general 
         // philosophy is that "assumptions are for bumpkins".
         trg[ZERO] = NULLCHAR ;
         byteCount = wcbsNOINSTALL ;
      }
   }

   #if WCB_DEBUG != 0
   if ( this->dbofs.is_open() )
   {
      this->gsdb.compose( "   '%s'\n   byteCount:%hd\n", trg, &byteCount ) ;
      this->dbofs << this->gsdb.ustr() << endl ;
   }
   #endif   // WCB_DEBUG

   return byteCount ;

}  //* End wcbGet() *

short WaylandCB::wcbGet ( gString& trg, bool primary, short len )
{
   trg.clear() ;                 // initialize caller's buffer

   char ctrg[gsMAXBYTES] ;
   short byteCount = this->wcbGet ( ctrg, primary, len ) ;
   if ( byteCount >= ZERO )
      trg = ctrg ;
   return byteCount ;

}  //* End wcbGet() *

short WaylandCB::wcbGet ( wchar_t* trg, bool primary, short len )
{
   short charCount = wcbsNOCONNECT ; // return value
   *trg = NULLCHAR ;                 // initialize caller's buffer

   if ( len <= gsMAXCHARS )
   {
      char ctrg[gsMAXBYTES] ;
      gString gs( trg ) ;        // convert to UTF-8
      gs.copy( ctrg, gsMAXBYTES ) ;
      short byteCount = this->wcbGet ( ctrg, primary, len ) ;
      if ( byteCount >= ZERO )
      {
         gs = ctrg ;
         gs.copy( trg, gsMAXCHARS ) ;
         charCount = gs.gschars() ;
      }
   }
   return charCount ;

}  //* End wcbGet() *

//*************************
//*        wcbSet         *
//*************************
//******************************************************************************
//* Public Method                                                              *
//* -------------                                                              *
//* Set contents of the Wayland clipboard.                                     *
//* Source data terminated at null terminator or if 'cnt' specified, then      *
//* specified number of items (PLUS null terminator if not included in count). *
//*                                                                            *
//* IMPORTANT NOTE: Always size your source buffer several bytes larger than   *
//*                 any data it will contain. This is to ensure that there is  *
//*                 space to "escape" any special characters that might be     *
//*                 misinterpreted by the shell program.                       *
//*                                                                            *
//* IMPORTANT NOTE: If the source data ('csrc') are wchar_t (UTF-32)           *
//*                 characters, then character input ('cnt') is limited to the *
//*                 width of the UTF-32 to UTF-8 conversion buffer(gsMAXBYTES).*
//*                 Therefore if the number of UTF-8 bytes required to         *
//*                 represent the UTF-32 source data is more than gsMAXBYTES,  *
//*                 truncation may occur.                                      *
//*                                                                            *
//* Input  : csrc    : pointer to buffer containing source data                *
//*                    wchar_t*  or  char*  or  gString&                       *
//*          primary : (optional, 'false' by default)                          *
//*                    'false', write to "regular" (ordinary, main) clipboard  *
//*                    'true', write to "primary" (highlighted-data) clipboard *
//*          cnt     : (optional, -1 by default)                               *
//*                    number of characters or bytes to write                  *
//*                    -- If default (-1) specified, write all source data to  *
//*                       and including the null terminator.                   *
//*                    -- If 'cnt' is reached before null terminator is        *
//*                       reached, a null terminator is appended and any       *
//*                       remaining source data are ignored.                   *
//*                    -- If 'cnt' == ZERO, then instead of processing 'csrc'  *
//*                       data, the wcbClear() method is called instead.       *
//*                                                                            *
//* Returns: for wchar_t source, number of characters written (incl.NULLCHAR)  *
//*          for char target or gString target, number of bytes written        *
//*            (incl. NULLCHAR)                                                *
//*          returns wcbsNOCONNECT if error i.e. not connected to clipboard    *
//*          returns wcbsNOINSTALL if 'wl-copy' utility not installed          *
//******************************************************************************
//* Notes:                                                                     *
//* 1) If 'wl-copy' is not installed, there will be a message simiar to:       *
//*         "sh: wl-copy: command not found"                                   *
//* 2) Note that double-quotes (and potentially other 'special' characters)    *
//*    in the source data must be escaped to avoid shell parsing errors.       *
//*                                                                            *
//*                                                                            *
//******************************************************************************

short WaylandCB::wcbSet ( const char* csrc, bool primary, short cnt )
{
   const char* const setTemplate = "wl-copy %s -t '%s' \"%s\" 1>>\"%s\" 2>>\"%s\"" ;
   //* "Special" characters" which must be escaped to prevent *
   //* the shell from misinterpreting them as commands.       *
   const char DQ = '"',
              DS = '$',
              BS = '\\',
              BT = '`' ;

   //* If not connected to the Wayland clipboard *
   //* interface, return an error condition.     *
   if ( ! this->connected )
      return ( wcbsNOCONNECT ) ;

   //* If 'cnt' == ZERO, it is assumed that *
   //* caller wants to clear the clipboard. *
   if ( cnt == ZERO )
      return ( this->wcbClear ( primary ) ) ;

   //* Discard contents (if any) of the target temp file.*
   ofstream ofs( this->tfspec, ofstream::out | ofstream::trunc ) ;
   if ( ofs.is_open() )
      ofs.close() ;

   //* Construct our local source buffer *
   if ( cnt < ZERO ) cnt = gsMAXBYTES ;   // if default buffer size specified
   char src[cnt + PADCNT] ;               // local buffer

   //* Determine the actual byte count of the souce data.*
   short byteCount = ZERO,    // return value
         specChars = ZERO ;   // number of escaped "special characters" found in source
   for ( byteCount = ZERO ; byteCount < cnt ; ++byteCount )
   {
      //* The backslash character ('\') IS the escape  *
      //* character, so special processing is required.*
      if ( csrc[byteCount] == BS )
      {
         //* Test whether an already-escaped character follows.*
         if ( (csrc[byteCount + 1] == DQ) || (csrc[byteCount + 1] == DS) ||
              (csrc[byteCount + 1] == BS) || (csrc[byteCount + 1] == BT) )
            ;  // do nothing
         else
            src[byteCount + specChars++] = BS ;
      }
      //* Test for other "special" characters.*
      else if ( (csrc[byteCount] == DQ) || (csrc[byteCount] == DS) ||
                (csrc[byteCount] == BT) )
      {
         src[byteCount + specChars++] = BS ;
      }
      src[byteCount + specChars] = csrc[byteCount] ;
      if ( src[byteCount + specChars] == NULLCHAR )
      {
         ++byteCount ;
         break ;
      }
      else if ( byteCount == (cnt - 1) ) // null terminator not found
      {
         src[byteCount + specChars] = NULLCHAR ;
         ++byteCount ;
         break ;
      }
   }

   //* Determine the size of the command buffer *
   gString gs( setTemplate ) ;
   short cmdLen = gs.utfbytes() + dfltMIME_LEN + (this->tfspecLen * 2),
         buffLen = cmdLen + byteCount + specChars + PADCNT ;
   char cmdBuff[buffLen] ;

   snprintf ( cmdBuff, buffLen, setTemplate, 
              (char*)(primary ? "-p" : " "), dfltMIME, src, this->tfspec, this->tfspec ) ;

   #if WCB_DEBUG != 0
   if ( this->dbofs.is_open() )
   {
      this->gsdb.compose( "-- wcbSet( buffLen:%hd )\n   %s\n   byteCount:%hd", 
                          &buffLen, cmdBuff, &byteCount ) ;
      this->dbofs << this->gsdb.ustr() << endl ;
   }
   #endif   // WCB_DEBUG

   this->Systemcall ( cmdBuff ) ;

   //* A successful call will produce no output to stdin/stderr. *
   //* If output generated, we can be relatively certain that an *
   //* error has occurred, but just in case, we read the first   *
   //* line of the file and test for tell-tale output.           *
   ifstream ifs( this->tfspec, ifstream::in ) ; // open the file
   if ( ifs.is_open() )                         // if file opened
   {
      cmdBuff[ZERO] = NULLCHAR ;
      ifs.getline( cmdBuff, buffLen, NEWLINE ) ;
      if ( ifs.good() || ifs.gcount() > ZERO )
      {
         gs = cmdBuff ;
         if ( ((gs.find( L"wl-copy" )) >= ZERO ) &&
              ((gs.find( L"command not found" )) >= ZERO) )
         {
            byteCount = wcbsNOINSTALL ;

            #if WCB_DEBUG != 0
            if ( this->dbofs.is_open() )
            {
               gs.erase( L'\n' ) ;
               this->gsdb.compose( "   wcbSet err: '%S'", gs.gstr() ) ;
               this->dbofs << this->gsdb.ustr() << endl ;
            }
            #endif   // WCB_DEBUG
         }
      }
      ifs.close() ;                             // close the file
   }
   return byteCount ;

}  //* End wcbSet() *

short WaylandCB::wcbSet ( const gString& src, bool primary, short cnt )
{

   return ( (this->wcbSet ( src.ustr(), primary, cnt )) ) ;

}  //* End wcbSet() *

short WaylandCB::wcbSet ( const wchar_t* src, bool primary, short cnt )
{
   //* Byte count is limited to width of gString object.*
   gString gs( src ) ;
   cnt = gs.utfbytes() ;

   short byteCount = this->wcbSet ( gs.ustr(), primary, cnt ),
         charCount = wcbsNOCONNECT ;
   if ( byteCount >= ZERO )
      charCount = gs.gschars() ;
   else
      charCount = byteCount ;

   return charCount ;

}  //* End wcbSet() *

//*************************
//*       wcbClear        *
//*************************
//******************************************************************************
//* Public Method:                                                             *
//* --------------                                                             *
//* Clear the clipboard.                                                       *
//*                                                                            *
//* Important Note: As detailed above, the wl-copy --clear ('-c') arguments    *
//* apparently do nothing. For this reason, by default we transmit an empty    *
//* string to clear the clipboard. The optional "useCmd" parameter is provided *
//* to allow testing of the --clear (-c) arguments.                            *
//*                                                                            *
//* Input  : primary: (optional, 'false' by default)                           *
//*                   if 'false', clear the "Regular" clipboard                *
//*                   if 'true',  clear the "Primary" clipboard                *
//*          useCmd : (optional, 'false' by default)                           *
//*                   if 'false', clear clipboard by sending an empty string   *
//*                   if 'true', _attempt_ to clear the clipboard using the    *
//*                              "--clear" argument                            *
//*                                                                            *
//* Returns: 'true' if successful                                              *
//*          'false' if not connected, or other error                          *
//******************************************************************************

bool WaylandCB::wcbClear ( bool primary, bool useCmd )
{
   bool status = false ;      // return value

   if ( this->connected )
   {
      //* It is assumed that the command will fit into a gString object.*
      gString gsCmd ;

      if ( ! useCmd )
      {
         if ( (this->wcbSet ( "", primary )) >= ZERO )
            status = true ;
      }
      else        // see note above
      {
         const char* const clrTemplate = 
                     "wl-copy %s --clear 1>>\"%s\" 2>>\"%s\"" ;
         gString gsCmd( clrTemplate, (primary ? "-p" : " "),
                        this->tfspec, this->tfspec ) ;

         //* Execute the command *
         this->Systemcall ( gsCmd.ustr() ) ;

         //* Verify that target clipboard is clear.*
         //* Only the nullchar should be returned. *
         char buff[gsMAXBYTES] ;
         if ( (this->wcbGet ( buff, primary )) == 1 )
            status = true ;
      }

      #if WCB_DEBUG != 0
      if ( this->dbofs.is_open() )
      {
         this->gsdb.compose( "   -- wcbClear(p:%hhd u:%hhd - %s)", 
                             &primary, &useCmd, (status != false ? "OK" : "ERROR") ) ;
         if ( (gsCmd.gschars()) > 1 )
            this->gsdb.append( "\n      %S", gsCmd.gstr() ) ;
         this->dbofs << this->gsdb.ustr() << endl ;
      }
      #endif   // WCB_DEBUG
   }
   return status ;

}  //* End wcbClear() *

//*************************
//*       wcbBytes        *
//*************************
//******************************************************************************
//* Public Method:                                                             *
//* --------------                                                             *
//* Report number of bytes of data stored on the specified clipboard.          *
//*                                                                            *
//* Input  : primary: (optional, 'false' by default)                           *
//*                   if 'false', report the "Regular" clipboard               *
//*                   if 'true',  report the "Primary" clipboard               *
//*                                                                            *
//* Returns: bytes of data on the clipboard incl. NULLCHAR                     *
//*          returns wcbsNOCONNECT if error i.e. not connected to clipboard    *
//*          returns wcbsNOINSTALL if 'wl-copy' utility not installed          *
//******************************************************************************
//* Note: If clipboard contains more than our buffer will hold, the value      *
//*       returned will be an undercount.                                      *
//******************************************************************************

short WaylandCB::wcbBytes ( bool primary )
{
   char* buff = new char[gsMAXBYTES * 4] ;

   short cbBytes = this->wcbGet ( buff, primary, (gsMAXBYTES * 4) ) ;
   delete [] buff ;

   return cbBytes ;

}  //* End wcbBytes() *

//*************************
//*       wcbTypes        *
//*************************
//******************************************************************************
//* Public Method:                                                             *
//* --------------                                                             *
//* Report the available formats for retrieving data from the specified        *
//* clipboard.                                                                 *
//*                                                                            *
//* Input  : trg    : buffer to receive descriptive text for available formats *
//*          cnt    : (optional, -1 by default) size of target buffer          *
//*                   for char* and gString targets, default: gsMAXBYTES       *
//*                   for wchar_t* target default: gsMAXCHARS                  *
//*                   NOTE: On average, approximately 200-300 bytes of (ASCII) *
//*                         data will be returned, but be safe.                *
//*          primary: (optional, 'false' by default)                           *
//*                   if 'false', report the "Regular" clipboard               *
//*                   if 'true',  report the "Primary" clipboard               *
//*                                                                            *
//* Returns: for wchar_t target, number of characters read (incl.NULLCHAR)     *
//*          for char target or gString target, number of bytes read           *
//*           (incl. NULLCHAR)                                                 *
//*          returns wcbsNOCONNECT if error i.e. not connected to clipboard    *
//*          returns wcbsNOINSTALL if 'wl-paste' utility not installed         *
//******************************************************************************

short WaylandCB::wcbTypes ( gString& trg, short cnt, bool primary )
{
   const char* const fmtTemplate = "wl-paste %s -nl 1>>\"%s\" 2>>\"%s\"" ;
   char buff[gsMAXBYTES] ;    // input buffer
   short byteCount = ZERO ;   // return value

   trg.clear() ;             // initialize caller's buffer

   //* If not connected to the Wayland clipboard *
   //* interface, return an error condition.     *
   if ( ! this->connected )
      return ( (-1) ) ;

   //* Discard contents (if any) of the target temp file.*
   ofstream ofs( this->tfspec, ofstream::out | ofstream::trunc ) ;
   if ( ofs.is_open() )
      ofs.close() ;

   //* Determine the size of the command buffer *
   gString gs( fmtTemplate ) ;
   short buffLen = gs.utfbytes() + (this->tfspecLen * 2) + PADCNT ;
   char cmdBuff[buffLen] ;

   //* Construct the command *
   snprintf ( cmdBuff, buffLen, fmtTemplate, 
              (char*)(primary ? "-p" : " "), this->tfspec, this->tfspec ) ;

   #if WCB_DEBUG != 0
   if ( this->dbofs.is_open() )
   {
      this->gsdb.compose( "-- wcbTypes( buffLen:%hd )\n   %s", &buffLen, cmdBuff ) ;
      this->dbofs << this->gsdb.ustr() << endl ;
   }
   #endif   // WCB_DEBUG

   //* Capture the clipboard *
   this->Systemcall ( cmdBuff ) ;

   //* Retrieve the results from the temp file.*
   ifstream ifs( this->tfspec, ifstream::in ) ; // open the file
   if ( ifs.is_open() )                         // if file opened
   {
      char inByte ;                             // input buffer
      while ( byteCount < (gsMAXBYTES - 1) )    // loop
      {
         ifs.read( &inByte, 1 ) ;               // read a byte
         if ( ifs.good() || (ifs.gcount() == 1) ) // if read successful
            buff[byteCount++] = inByte ;        // save the byte
         else                                   // end-of-file
            break ;
      }
      //* Discard any trailing newline and terminate the string.*
      if ( (byteCount > ZERO) && (buff[byteCount - 1] == NEWLINE) )
         --byteCount ;
      if ( (byteCount == ZERO) || (buff[byteCount] != NULLCHAR) )
         buff[byteCount++] = NULLCHAR ;
      ifs.close() ;                             // close the file
   }

   #if WCB_DEBUG != 0
   if ( this->dbofs.is_open() )
   {
      this->gsdb.compose( "   '%s'\n   byteCount:%hd\n", buff, &byteCount ) ;
      this->dbofs << this->gsdb.ustr() << endl ;
   }
   #endif   // WCB_DEBUG

   trg = buff ;

   //* Test for a "wl-paste not found" message.*
   if ( ((trg.find( L"wl-paste" )) >= ZERO ) &&
        ((trg.find( L"command not found" )) >= ZERO) )
   {
      trg.clear() ;
      byteCount = wcbsNOINSTALL ;
   }

   return ( byteCount ) ;

}  //* End wcbTypes() *

short WaylandCB::wcbTypes ( char* trg, short cnt, bool primary )
{
   gString gs ;
   this->wcbTypes ( gs, cnt, primary ) ;
   if ( cnt < ZERO )
      cnt = gsMAXBYTES ;
   gs.copy( trg, cnt ) ;
   gs = trg ;        // count the bytes
   short utfbytes = gs.utfbytes() ;

   return utfbytes ;

}  //* End wcbTypes() *

short WaylandCB::wcbTypes ( wchar_t* trg, short cnt, bool primary )
{
   gString gs ;
   this->wcbTypes ( gs, cnt, primary ) ;
   if ( cnt < ZERO )
      cnt = gsMAXCHARS ;
   gs.copy( trg, cnt ) ;
   gs = trg ;        // count the characters
   short wchars = gs.gschars() ;

   return wchars ;

}  //* End wcbTypes() *

//*************************
//*    wcbIsConnected     *
//*************************
//******************************************************************************
//* Public Method:                                                             *
//* --------------                                                             *
//* Report whether connection to the Wayland clipboard has been established.   *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: 'true' if connected to clipboard, else 'false'                    *
//******************************************************************************

bool WaylandCB::wcbIsConnected ( void )
{

   return ( this->connected ) ;

}  //* End wcbIsConnected() *

//*************************
//*        wcbTest        *
//*************************
//******************************************************************************
//* Public Method                                                              *
//* -------------                                                              *
//* Test the connection with the Wayland clipboard.                            *
//* Send a short (< 1000 bytes) message to the clipboard, then retrieve the    *
//* data and compare the results.                                              *
//*                                                                            *
//* Input  : testdata : a short null-terminated string to use as test data     *
//*                                                                            *
//* Returns: Member of enum wcbStatus:                                         *
//******************************************************************************

wcbStatus WaylandCB::wcbTest ( const char* testdata )
{
   gString gs ;
   wcbStatus status = wcbsNOCONNECT ;

   //* If currently connected, test the connection.*
   if ( this->connected )
   {
      if ( (this->wcbSet ( testdata )) >= ZERO )
      {
         if ( (this->wcbGet ( gs )) >= ZERO )
         {
            if ( (gs.compare( testdata )) == ZERO )
               status = wcbsACTIVE ;
         }
      }
   }

   //* Else, not connected, diagnose the connection.*
   else
   {
      short s = this->wcbTypes ( gs ) ; //(should never return zero)
      if ( s >= ZERO )
         status = wcbsACTIVE ;
      else if ( s == wcbsNOCONNECT )
         status = wcbsNOCONNECT ;
      else if ( s <= wcbsNOINSTALL )
         status = wcbsNOINSTALL ;
   }
   return status ;

}  //* End wcbTest() *

//*************************
//*       wcbReinit       *
//*************************
//******************************************************************************
//* Public Method                                                              *
//* -------------                                                              *
//* Terminate the current connection (if any) with the Wayland clipboard,      *
//* release all local resources, reinitialize local data members, and          *
//* re-establish the connection.                                               *
//*                                                                            *
//* The 'wl-copy' and/or 'wl-paste' utility will occasionally go off into the  *
//* weeds. The reasons for this are unclear, but communication can usually be  *
//* re-established by running the initialization sequence again. However, to   *
//* do this, it is necessary to first reset all class resources to the initial *
//* state and then run the startup sequence.                                   *
//*                                                                            *
//* Note that this method does not reference the 'connected' member, but       *
//* unconditionally attempts the reconnection.                                 *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: 'true' if connection re-established, else 'false'                 *
//******************************************************************************

bool WaylandCB::wcbReinit ( void )
{
   //* Delete the temporary file.*
   this->DeleteTempname ( this->tfspec ) ;

   //* Reset all data members EXCEPT for debugging members.*
   this->reset () ;

   //* Execute the startup sequence.*
   this->StartupSequence () ;

   return ( this->connected ) ;

}  //* End wcbReinit() *

//*************************
//*      wcbVersion       *
//*************************
//******************************************************************************
//* Public Method:                                                             *
//* --------------                                                             *
//* Report the WaylandCB-class version number.                                 *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: pointer to const string of the format: "x.y.zz"                   *
//******************************************************************************

const char* WaylandCB::wcbVersion ( void )
{

   return ( WaylandCB_Version ) ;

}  //* End wcbVersion() *

//*************************
//*         reset         *
//*************************
//******************************************************************************
//* Private Method:                                                            *
//* ---------------                                                            *
//* Set all data members to their initial state.                               *
//* Called only by constructors and by wcbReinitialize() method.               *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

void WaylandCB::reset ( void )
{
   *this->tfdir    = NULLCHAR ;
   *this->tfspec   = NULLCHAR ;
   this->tfspecLen = ZERO ;
   this->connected = false ;

}  //* End reset() *

//*************************
//*    StartupSequence    *
//*************************
//******************************************************************************
//* Private Method:                                                            *
//* ---------------                                                            *
//* Establish temp file for communications with the external "wl-clipboard"    *
//* utilities, "wl-copy" and "wl-paste".                                       *
//*                                                                            *
//* Establish the connection with the system clipboard.                        *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: nothing                                                           *
//*          'connected' data member set if successful                         *
//******************************************************************************

void WaylandCB::StartupSequence ( void )
{
   gString gs ;

   //* Locate the system's temp-file directory.*
   if ( (this->GetTempdirPath ( gs )) )
   {
      //* Save the temp-dir filespec.*
      gs.copy( this->tfdir, gsMAXBYTES ) ;

      //* Create the primary temp file.*
      if ( (this->CreateTempname ( gs )) )
      {
         gs.copy( this->tfspec, gsMAXBYTES ) ;
         this->tfspecLen = gs.utfbytes() ;

         //* Establish connection with wl-clipboard.*
         this->connected = this->EstablishIndirectConnection () ;
      }
   }

}  //* End StartupSequence() *

//*******************************
//* EstablishIndirectConnection *
//*******************************
//******************************************************************************
//* Private Method:                                                            *
//* ---------------                                                            *
//* Determine whether the "wl-copy" and "wl-paste" utilities are installed     *
//* on the system, and if they are, perform a simple copy-and-paste operation  *
//* to establish the connection.                                               *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: 'true' if utilities are installed and clipboards are accessible.  *
//*          'false' otherwise                                                 *
//******************************************************************************
//* The connection is looped for a maximum success rate.                       *
//* As noted elsewhere, the 'wl-copy' and 'wl-paste' utilities are slown and   *
//* fragile. Problems that mysteriously arise, just as mysteriously disappear. *
//* The looped sequence seems to be the best way to overcome the gremlins      *
//* in the wl-clipboard interface.                                             *
//*                                                                            *
//* Additional Note: Because the 'wl-copy' utility is poorly written and       *
//* fragile, if the application encounters and error during copy/paste then    *
//* the next time the application starts we may unable to establish the        *
//* connection. We need an elegant way to clear the error, so we can report    *
//* a successful connection.                                                   *
//*                                                                            *
//******************************************************************************

bool WaylandCB::EstablishIndirectConnection ( void )
{
   const char* const testMsg = "WaylandCB Test Message" ;
   chrono::duration<short, std::milli>aMoment( 6 ) ;  // 6-millisecond delay
   gString gs( "%s (primary)", testMsg ) ;
   char trg[gsMAXBYTES] ;
   short setStatus ;
   bool status = false ;

   this->connected = true ;   // temporarily allow access to the public methods

   #if WCB_DEBUG !=  0  // DEBUG ONLY
   this->dbofs << "-- EstablishIndirectConnection()" << endl ;
   #endif   // WCB_DEBUG

   //* Perform the copy-and-paste test. *
   //* Loop until success or exhaustion.*
   for ( short testCnt = 3 ; testCnt > ZERO ; --testCnt )
   {
      //* Place the test message on the "Primary" clipboard.*
      if ( (setStatus = this->wcbSet ( gs, true )) >= ZERO )
      {
         this_thread::sleep_for( aMoment ) ;    // wait a moment

         //* Retrieve the contents of the "Primary" clipboard.*
         if ( (this->wcbGet ( trg, true )) >= ZERO )
         {
            //* Compare the data returned with the data transmitted. *
            //* If a match, then connection is verified.             *
            if ( (gs.compare( trg, true )) == ZERO )
               status = true ;
   
            //* Delete the test message from the clipboard.*
            this->wcbClear ( true ) ;
            break ;  // success !!
         }
      }
      else if ( setStatus == wcbsNOINSTALL ) // 'wl-copy' not installed
         break ;

      //* If operation failed, wait a moment before retry.*
      this_thread::sleep_for( aMoment ) ;
   }

   this->connected = status ;    // remember the results of the test

   #if WCB_DEBUG !=  0  // DEBUG ONLY
   this->dbofs << "-- EstablishIndirectConnection(" 
               << (this->connected ? "OK" : "ERROR") ;
    if ( setStatus == wcbsNOINSTALL )
       dbofs << " - 'wl-copy' not installed" ;
    this->dbofs << ")\n" << endl ;
   #endif   // WCB_DEBUG

   return status ;

}  //* End EstablishIndirectConnection() *

//*************************
//*      Systemcall       *
//*************************
//******************************************************************************
//* Private Method:                                                            *
//* ---------------                                                            *
//* Send a command to the shell program.                                       *
//*                                                                            *
//* Input  : cmd :                                                             *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************
//* We execute a short delay before calling the system.                        *
//* This is done because the wl-clipboard utilities are outrageously slow      *
//* to respond. See notes in the module header.                                *
//*                                                                            *
//******************************************************************************

void WaylandCB::Systemcall ( const char* cmd )
{
   static chrono::duration<short, std::milli>aMoment( 50 ) ;

   this_thread::sleep_for( aMoment ) ;
   system ( cmd ) ;

}  //* End Systemcall() *

//*************************
//*    GetTempdirPath     *
//*************************
//******************************************************************************
//* Private Method:                                                            *
//* ---------------                                                            *
//* This is a stub function which provides access to the C++17 function:       *
//*                       temp_directory_path()                                *
//*                                                                            *
//* Because most programmer's will not reach C++17 for at least several        *
//* months, we isolate the call here and implement what we expect is the way   *
//* the library function will be implemented.                                  *
//*         (This WILL NOT work under Windoze--and we don't care.)             *
//*         (Screw Windoze, and the horse it rode in on.         )             *
//* -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- *
//* From <http://en.cppreference.com>                                          *
//* ---------------------------------                                          *
//* These functions are available in C++17 (released Jan. 2018):               *
//* const char* fs::temp_directory_path ( void ) ;                             *
//* const char* fs::temp_directory_path ( std::error_code& ec ) ;              *
//*                                                                            *
//* Returns A directory suitable for temporary files. The path is guaranteed   *
//* to exist and to be a directory. The overload that takes error_code&        *
//* argument returns an empty path on error.                                   *
//*                                                                            *
//* Exceptions: The overload that does not take a std::error_code& parameter   *
//* throws filesystem_error on underlying OS API errors, constructed with path *
//* to be returned as the first argument and the OS error code as the error    *
//* code argument. std::bad_alloc may be thrown if memory allocation fails.    *
//* The overload taking a std::error_code& parameter sets it to the OS API     *
//* error code if an OS API call fails, and executes ec.clear() if no errors   *
//* occur.                                                                     *
//*                                                                            *
//* On POSIX systems, the path may be the one specified in the environment     *
//* variables TMPDIR, TMP, TEMP, TEMPDIR, and, if none of them are specified,  *
//* the path "/tmp" is returned.                                               *
//*                                                                            *
//* Input  : tdPath : (by reference) receives the path string                  *
//*                                                                            *
//* Returns: 'true'  if successful, 'false' if system error                    *
//******************************************************************************

bool WaylandCB::GetTempdirPath ( gString& tdPath )
{
   //* Default path of temp directory on GNU/Linux systems. *
   //* (Used only if environment variable not set.)         *
   const char* const dfltPath = "/tmp" ;

   const char* envPath ;            // returned by getenv()
   FileStats   fStats ;             // target filestats
   bool        status = false ;     // return value

   if ( (envPath = std::getenv ( "TMPDIR" )) == NULL )
      if ( (envPath = std::getenv ( "TMP" )) == NULL )
         if ( (envPath = std::getenv ( "TEMP" )) == NULL )
            if ( (envPath = std::getenv ( "TEMPDIR" )) == NULL )
               envPath = dfltPath ;
   tdPath = envPath ;
   if ( (stat64 ( tdPath.ustr(), &fStats )) == ZERO )
   {
      if ( ((S_ISDIR(fStats.st_mode)) != false) &&
           ((access ( tdPath.ustr(), R_OK )) == ZERO) && 
           ((access ( tdPath.ustr(), W_OK )) == ZERO) )
         status = true ;
   }
   return status ;

}  //* End GetTempdirPath()

//*************************
//*    CreateTempname     *
//*************************
//******************************************************************************
//* Private Method:                                                            *
//* ---------------                                                            *
//* Create a unique path/filename for a temporary file.                        *
//*                                                                            *
//* Input  : tmpPath: (by reference) receives the path/filename                *
//*                                                                            *
//* Returns: 'true'  if path/filename created successfully                     *
//*          'false' if library call failed                                    *
//******************************************************************************

bool WaylandCB::CreateTempname ( gString& tmpPath )
{
   bool  status = false ;

   char tn[gsMAXBYTES] ;
   gString gstmp( "%s/WLCB_XXXXXX", this->tfdir ) ;
   gstmp.copy( tn, gsMAXBYTES ) ;
   int descriptor ;
   if ( (descriptor = mkstemp ( tn )) != (-1) )
   {
      close ( descriptor ) ;     // close the file
      tmpPath = tn ;             // copy target path to caller's buffer
      status = true ;            // declare success
   }
   else
      tmpPath.clear() ;
   return status ;

}  //* End CreateTempname() *

//*************************
//*    DeleteTempname     *
//*************************
//******************************************************************************
//* Private Method:                                                            *
//* ---------------                                                            *
//* Delete the specified temporary file.                                       *
//*                                                                            *
//* Input  : tmpPath: (by reference) filespec of temp file to be deleted       *
//*                                                                            *
//* Returns: 'true'  if file deleted                                           *
//*          'false' if library call failed (file not found, or system error)  *
//******************************************************************************

bool WaylandCB::DeleteTempname ( const gString& tmpPath )
{
   bool  status = false ;

   if ( (unlink ( tmpPath.ustr())) == ZERO )
      status = true ;

   return status ;

}  //* End DeleteTempname() *

#if 0    // NOT CURRENTLY USED
//* Specifies the character or group of characters to be "escaped" *
//* by a call to the EscapeSpecialChars() method.                  *
enum escGroup : short
{
   escDQ,      // escape only double-quote character ( " )
   escMIN,     // escape the four most critical characters: ( $ \ ` " )
   escLINUX,   // escMIN plus the Linux shell special characters ( \ ' " ? : ; & > < | * )
   escMAX,     // escLINUX plus all remaining special characters
               // # + = % ! ~ { } @ ( ) [ ] and space(' ' 20h)
   escWCHAR    // escape only the specified character
} ;

static short EscapeSpecialChars ( gString& src, escGroup group, 
                                  wchar_t wch = NULLCHAR ) ;
static short EscapeSpecialChars ( wchar_* src, escGroup group, 
                                  wchar_t wch = NULLCHAR ) ;
static short EscapeSpecialChars ( char* src, escGroup group, 
                                  char ch = NULLCHAR ) ;

//*************************
//*  EscapeSpecialChars   *
//*************************
//******************************************************************************
//* Scan the source string (typically a filespec). If the string contains any  *
//* characters of the specified character group (enum escGroup) 'escape' them  *
//* with a backslash character.                                                *
//*                                                                            *
//* Do not use this method for URL (Universal Resource Locator) specs.         *
//* URL specifications require a specific protocol which is not handled here.  *
//*                                                                            *
//* Input  : pathSpec : (by reference) source text data                        *
//*                     on return the data has bee 'escaped' accoring to the   *
//*                     specified criteria                                     *
//*          group    : (member of enum escGroup) specifies the group of       *
//*                     characters to be 'escaped'                             *
//*          wch      : (optional, NULLCHAR by default)                        *
//*                     if 'group' parameter == escWCHAR, then 'wch' specifies *
//*                     a single character (NOT nullchar) to be 'escaped'      *
//*                                                                            *
//* Returns: number of characters 'escaped'                                    *
//******************************************************************************
//* Special Case Testing for '\' as the special character:                     *
//* 1) If the incomming data contains a '\', we must test whether:             *
//*    a) The '\' itself has already been escaped,                             *
//*    b) The '\' is escaping some other special character,                    *
//*    c) The '\' is a special character which needs to be escaped.            *
//* 2) If data contains an escaped special character, but caller intended      *
//*    for it to be two SEPERATE special characters, caller will be            *
//*    disappointed that we could not read his/her/its mind. (sorry about that)*
//*                                                                            *
//* -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -  *
//* For more information, see "Quoting Special Characters in BASH.odt".        *
//******************************************************************************

static short EscapeSpecialChars ( gString& pathSpec, escGroup group, wchar_t wch )
{
   //* List of critical "special" characters.*
   const short   minSpecialCount = 4 ;
   const wchar_t minSpecialChars[minSpecialCount] = 
   { L'\\', L'"', L'`', L'$' } ;

   //* List of Linux/UNIX shell "special" characters.*
   const short   linuxSpecialCount = 13 ;
   const wchar_t linuxSpecialChars[linuxSpecialCount] = 
   { L'\\', L'"', L'\'', L'?', L':', L';', L'&', L'>', L'<', L'|', L'*', L'`', L'$' } ;

   //* List of all known "special" characters which might be interpreted *
   //* by the shell or shell programs as commands rather than plain text.*
   const short   maxSpecialCount = 27 ;
   const wchar_t maxSpecialChars[maxSpecialCount] = 
   { L'\\', L'"', L'\'', L'?', L':', L';', L'&', L'>', L'<', L'|', L'*', L'`', L'$',
     L'#', L'+', L'=', L'%', L'!', L'~', L'{', L'}', L'@', L'(', L')', L'[', L']', L' ' } ;

   const wchar_t wDQ = L'"',
                 wBS = L'\\' ;

   const wchar_t *wPtr ;            // pointer to wchar_t array
   short wCount,                    // number of elements in wchar_t array
         indx,                      // text index
         specChars = ZERO ;         // return value

   //* Point to the reference array and establish number of array elements.*
   if ( group == escMIN )
   { wPtr = minSpecialChars ;   wCount = minSpecialCount ; }
   else if ( group == escLINUX )
   { wPtr = linuxSpecialChars ; wCount = linuxSpecialCount ; }
   else if ( group == escMAX )
   { wPtr = maxSpecialChars ;   wCount = maxSpecialCount ; }
   else if ( group == escDQ )
   { wPtr = &wDQ ;              wCount = 1 ; }
   else if ( group == escWCHAR )
   {  //* Do not allow escaping of null terminator.*
      wPtr = &wch ;
      wCount = (wch != NULLCHAR) ? 1 : 0 ;
   }

   for ( short specIndx = ZERO ; specIndx < wCount ; ++specIndx )
   {
      indx = ZERO ;
      do
      {
         if ( (indx = pathSpec.find( wPtr[specIndx], indx )) >= ZERO )
         {
            //* Special character == '\'. (see note above) *
            if ( (pathSpec.gstr()[indx]) == wBS )
            {
               bool insert = true ;
               //* If escape character is itself escaped *
               if ( (indx > ZERO) && ((pathSpec.gstr()[indx - 1]) == wBS) )
               { ++indx ; insert = false ; }
               //* If '\' is being used as an escape character *
               if ( insert )
               {
                  for ( short i = ZERO ; i < wCount && insert ; ++i )
                  {
                     if ( (pathSpec.gstr()[indx + 1]) == wPtr[i] )
                     { indx += 2 ; insert = false ; }
                  }
               }

               if ( insert )  // escape the escape character
               {
                  pathSpec.insert( wBS, indx ) ;
                  ++specChars ;
                  indx += 2 ;
               }
               else
                  ++indx ;
            }
            else
            {
               if ( indx == ZERO )       // first character in array
               {
                  pathSpec.insert( wBS, indx ) ;
                  ++specChars ;
                  indx += 2 ;
               }
               else // (indx > ZERO), second or later character in array
               {
                  if ( (pathSpec.gstr()[indx - 1]) != wBS )
                  {
                     pathSpec.insert( wBS, indx ) ;
                     ++specChars ;
                     indx += 2 ;
                  }
                  else
                     ++indx ;
               }
            }
         }
      }
      while ( (indx >= ZERO) && (pathSpec.gstr()[indx] != NULLCHAR) ) ;
   }

   return specChars ;

}  //* End EscapeSpecialChars() *
#endif   // NOT CURRENTLY USED

//*************************
//*                       *
//*************************
//******************************************************************************
//*                                                                            *
//*                                                                            *
//*                                                                            *
//* Input  :                                                                   *
//*                                                                            *
//* Returns:                                                                   *
//******************************************************************************

